Worker

Struct Worker 

pub struct Worker<'q, T: Task> { /* private fields */ }
A thread-local view of the task queue.

This is the primary interface to takeaway. It manages a set of pending tasks, communicating with the global Queue to distribute tasks across the system, and provides methods for adding and retrieving tasks.

Usage

You can obtain a Worker from Worker::new(). To enqueue tasks, use Worker::enqueue(). To retrieve tasks, use Worker::next(). In either case, lower-level interfaces are provided for more control.

// Construct the 'Worker'.
let mut worker = Worker::new(queue, id);

// Enqueue some initial tasks.
worker.enqueue_one(todo!());

// Process tasks.
while let Some(task) = worker.next().await {
    // Execute the task.
    //...

    // Enqueue sub-tasks as necessary.
    worker.enqueue_one(todo!());
}

// The task queue has shut down.

Implementations

Source

impl<'q, T: Task> Worker<'q, T>

pub fn new(queue: &'q Queue<T>, id: usize) -> Self

Construct a new Worker.

The worker will be assigned the specified ID.

Panics
  • Panics if id >= queue.config().num_workers().get().
  • Panics if another worker has been assigned this ID already.
pub const fn id(&self) -> usize

The ID of this worker.

This is the value assigned to the worker in Worker::new().

pub const fn queue(&self) -> &'q Queue<T>

The associated global queue.

pub const fn enqueuer(&self) -> &Rc<Enqueuer<T>>

The associated Enqueuer.

pub fn enqueue(&self, tasks: impl IntoIterator<Item = T>)

Enqueue a set of tasks.

The tasks will be added to a thread-local Vec, where they will remain until they are used in the Worker’s batch.

This is a shorthand for self.enqueuer().extend(tasks). If the worker is borrowed mutably, the Enqueuer can be used directly by cloning the Rc returned by Worker::enqueuer().

pub fn enqueue_one(&self, task: T)

Enqueue a single task.

The task will be added to a thread-local Vec, where it will remain until it is used in the Worker’s batch.

This is a shorthand for self.enqueuer().add(task).

pub async fn next(&mut self) -> Option<T>

Retrieve a single task asynchronously.

If None is returned, the task queue has been shut down.

pub fn poll(&mut self, waker: &Waker) -> Poll<Option<T>>

Poll for a single task.

If None is returned, the task queue has been shut down.